/*
 * The Gold Partition Conjecture for K-thin posets
 *
 * gpc1.c
 * Main module
 */

/* Copyright 2005-2006, Marcin Peczarski. All Rights Reserved.
 *
 * This program is free software. You may use this software, and copy,
 * redistribute, and/or modify it for non profit works.
 * Please do not remove this copyright notice.
 */

#include "gpc.h"

static void main_init(void);
static void main_clean(void);

int in_file = -1, out_file = -1;
extern clock_t start_time;

/*
 * Global data initialization and deinitialization
 */

static void main_init(void) {
  gpc2_module_init();
  poset_module_init();
}

static void main_clean(void) {
  if (in_file >= 0)
    close(in_file);
  if (out_file >= 0)
    close(out_file);
  gpc2_module_clean();
  poset_module_clean();
}

/*
 * Program administration
 */

void fatal_error(const char *fmt, ...) {
  va_list fmt_args;

  fprintf(stderr, "ERROR: ");
  va_start(fmt_args, fmt);
  vfprintf(stderr, fmt, fmt_args);
  va_end(fmt_args);
  if (errno) {
    fprintf(stderr, ", errno=%d", errno);
    perror(" ");
  }
  else
    fprintf(stderr, "\n");
  main_clean();
  exit(2);
}

int main(int argc, char *argv[]) {
  long offset, count;
  int  action, result = 0;

  start_time = times(NULL);
  if (argc < 2) {
    printf("usage: %s action [parameters]\ncompiled at "__DATE__
           " "__TIME__"\n", argv[0]);
    #ifdef _INT_INDEX_
    printf("int indices\n");
    #else
    printf("long indices\n");
    #endif
    #ifdef _PRINT_LOG_
    printf("printing log on\n");
    #else
    printf("printing log off\n");
    #endif
    #ifdef _CHECK_CORRECTNESS_
    printf("checking code included\n");
    #else
    printf("checking code not included\n");
    #endif
    #if defined _USE_MIN_ELEMENT_TABLE_
      printf("min element table\n");
    #elif defined _USE_BUILT_IN_MIN_ELEMENT_
      printf("built-in min element\n");
    #endif
    #if defined _USE_BUILT_IN_MAX_ELEMENT_
      printf("built-in max element\n");
    #endif
    printf("THINNES = %d   VERSION = %d\n"
           "MAX_STACK = %ld   HASH_TBL_SIZE = %ld\n"
           "MAX_COMPONENTS = %d   MAX_PERM = %d\n",
           (int)K, (int)VERSION,
           (long)MAX_STACK, (long)HASH_TBL_SIZE,
           (int)MAX_COMPONENTS, (int)MAX_PERM);
    #if defined _CPCP_CACHE_
      printf("CPCP_CACHE_SIZE = 0x%X\n", CACHE_SIZE);
    #endif
    printf("\n");
    return 3;
  }
  action = atoi(argv[1]);
  switch (action) {

    case 0: /* Create initial file containing the first configuration. */
      if (argc < 3) {
        printf("usage: %s 0 out_file\n\n", argv[0]);
        return 3;
      }
      out_file = open(argv[2], O_RDWR | O_CREAT | O_TRUNC, FATR);
      if (out_file < 0)
        fatal_error("can't create file %s", argv[2]);
      first_configuration();
      return 0;

    case 1: /* Make one step. */
      if (argc < 4) {
        printf("usage: %s 1 in_file out_file [offset [count]]\n\n",
               argv[0]);
        return 3;
      }
      offset = argc >= 5 ? atol(argv[4]) : 0;
      count  = argc >= 6 ? atol(argv[5]) : 0;
      in_file = open(argv[2], O_RDONLY);
      if (in_file < 0)
        fatal_error("can't open file %s", argv[2]);
      out_file = open(argv[3], O_RDWR | O_CREAT | O_EXCL, FATR);
      if (out_file < 0 && errno == EEXIST) {
        out_file = open(argv[3], O_RDWR);
        if (out_file < 0)
          fatal_error("can't open file %s", argv[3]);
        else {
          errno = 0;
          main_init();
          result = one_step(TRUE, offset, count);
        }
      }
      else if (out_file < 0)
        fatal_error("can't create file %s", argv[3]);
      else {
        main_init();
        result = one_step(FALSE, offset, count);
      }
      main_clean();
      return result;

    case 5: /* Display overview of a file. */
    case 6: /* Display contents of a file. */
      if (argc < 3) {
        printf("usage: %s %d files_list\n\n", argv[0], action);
        return 3;
      }
      main_init();
      print_files_contents(argc, argv, action == 6);
      main_clean();
      return 0;

    default:
      printf("unknown action: %s\n", argv[1]);
      return 3;
  }
}

/*
 * Auxiliary routines
 */

void empty_file_header(FILE_HEADER* fh) {
  memset(fh, 0, sizeof(FILE_HEADER));
  fh->thinness          = K;
  fh->version           = VERSION;
  fh->step              = 0;
  fh->max_elements      = 0;
  fh->file_length       = sizeof(FILE_HEADER);
  fh->input_offset      = 0;
  fh->input_posets      = 0;
  fh->output_posets     = 0;
  fh->considered_posets = 0;
  fh->search_calls      = 0;
  fh->isomorphism_calls = 0;
  fh->exhaustive_calls  = 0;
  fh->cpcp_calls        = 0;
  fh->combinations      = 0;
  fh->hash_tbl_usage    = 0;
  fh->stack_usage       = 0;
  fh->real_time         = 0;
  fh->user_time         = 0;
  fh->syst_time         = 0;
  fh->max_extensions    = 0;
  fh->max_components    = 0;
  fh->max_permutations  = 0;
}

void compare_file_headers(FILE_HEADER* fh0, const FILE_HEADER* fh1) {
  if (fh0->step < 0)
    fh0->step = fh1->step;
  else if (fh0->step != fh1->step)
    fatal_error("compare_file_headers, files belong to different iterations");
  if (fh0->input_offset     > fh1->input_offset)
      fh0->input_offset     = fh1->input_offset;
  if (fh0->max_elements     < fh1->max_elements)
      fh0->max_elements     = fh1->max_elements;
  if (fh0->hash_tbl_usage   < fh1->hash_tbl_usage)
      fh0->hash_tbl_usage   = fh1->hash_tbl_usage;
  if (fh0->stack_usage      < fh1->stack_usage)
      fh0->stack_usage      = fh1->stack_usage;
  if (fh0->max_extensions   < fh1->max_extensions)
      fh0->max_extensions   = fh1->max_extensions;
  if (fh0->max_components   < fh1->max_components)
      fh0->max_components   = fh1->max_components;
  if (fh0->max_permutations < fh1->max_permutations)
      fh0->max_permutations = fh1->max_permutations;
  fh0->input_posets      += fh1->input_posets;
  fh0->output_posets     += fh1->output_posets;
  fh0->considered_posets += fh1->considered_posets;
  fh0->search_calls      += fh1->search_calls;
  fh0->isomorphism_calls += fh1->isomorphism_calls;
  fh0->exhaustive_calls  += fh1->exhaustive_calls;
  fh0->cpcp_calls        += fh1->cpcp_calls;
  fh0->combinations      += fh1->combinations;
  fh0->real_time         += fh1->real_time;
  fh0->user_time         += fh1->user_time;
  fh0->syst_time         += fh1->syst_time;
  fh0->file_length       += fh1->file_length;
}

static void print_ticks(const char* s, clock_t t, long c) {
  double d = (double)t / (double)c;

  printf("%-12s %11.3f\n", s, d);
}

void print_file_header(const FILE_HEADER* fh) {
  long clk_tck;
  char cpc[24];

  clk_tck = sysconf(_SC_CLK_TCK);
  sprintf(cpc, CPC_PRINT_FORMAT, fh->max_extensions);
  printf("thinness %d  iteration %d  file length %ld\n"
         "input offset              %12ld\n"
         "input configurations      %12ld\n"
         "output configurations     %12ld\n"
         "considered configurations %12ld\n"
         "search calls              %12ld\n"
         "isomorphism calls         %12ld\n"
         "exhaustive calls          %12ld\n"
         "counting extensions calls %12ld\n"
         "combinations checked      %12ld\n"
         "hash table usage          %12ld\n"
         "stack usage               %12ld\n"
         "max elements              %12d\n"
         "max components            %12d\n"
         "max permutations          %12d\n"
         "max extensions %23.23s\n",
         fh->thinness, fh->step, fh->file_length, fh->input_offset,
         fh->input_posets, fh->output_posets, fh->considered_posets,
         fh->search_calls, fh->isomorphism_calls, fh->exhaustive_calls,
         fh->cpcp_calls, fh->combinations, fh->hash_tbl_usage, fh->stack_usage,
         fh->max_elements, fh->max_components, fh->max_permutations, cpc);
  print_ticks("real time", fh->real_time, clk_tck);
  print_ticks("user time", fh->user_time, clk_tck);
  print_ticks("system time", fh->syst_time, clk_tck);
}
