/*
 * The Gold Partition Conjecture for K-thin posets
 *
 * gpc2.c
 * Proof routines
 */

/* Copyright 2005-2006, Marcin Peczarski. All Rights Reserved.
 *
 * This program is free software. You may use this software, and copy,
 * redistribute, and/or modify it for non profit works.
 * Please do not remove this copyright notice.
 */

#include "gpc.h"

#define EMPTY_POS -1 /* the indicator of an empty position in the hash table */
extern int     in_file, out_file;
extern long    isomorphic_calls_number, exhaustive_calls_number;
extern long    cpcp_calls_number;
static INDEX   h1_tbl[N][N]; /* table speed up counting the hash function */
static INDEX   new_posets_count;
static INDEX   hash_tbl[HASH_TBL_SIZE];
static POSET   posets_stack[MAX_STACK];
static POSET   pxyz[MAX_COMPONENTS + 1];
static int     n_components;
static CPC     txyz[MAX_COMPONENTS + 1], t2cache[MAX_COMPONENTS + 1][N][N];
static BOOL    cache_not_valid;
static int     lw2_x, lw2_y, lw3_x, lw3_y, lw3_z; /*previously found witness*/
static int     old_perm[MAX_PERM][N], new_perm[MAX_PERM][N];
static clock_t user_time = 0, syst_time = 0;
clock_t        start_time;

/*
 * Module initialization and deinitialization
 */

void gpc2_module_init(void) {
  INDEX w1[(N + 1) * (N + 1) / 4];
  int   i, m1, m2;

  w1[0] = 1;
  for (i = 1; i < (N + 1) * (N + 1) / 4; i++)
    w1[i] = (w1[i - 1] * N) % HASH_TBL_SIZE;
  for (m2 = 0; m2 < N; m2++)
    for (m1 = 0; m1 <= m2; m1++)
      h1_tbl[m1][m2] = h1_tbl[m2][m1] = w1[m1 * (N - m1) + m2];
}

void gpc2_module_clean(void) {
}

/*
 * Input/output routines
 */

static void get_poset(void) {
  size_t l;
  int    i, j;
  char   t[MAX_PERM * N];

  if (read(in_file, posets_stack, sizeof(POSET)) != sizeof(POSET))
    fatal_error("get_poset, reading of the poset failed");
  l = posets_stack[0].p * posets_stack[0].m;
  if (read(in_file, t, l) != l)
    fatal_error("get_poset, reading of permutations failed");
  for (i = 0; i < posets_stack[0].p; i++) {
    for (j = 0; j < posets_stack[0].m; j++)
      old_perm[i][j] = t[i * posets_stack[0].m + j];
    for (; j < N; j++)
      old_perm[i][j] = j;
  }
}

static long put_poset(const POSET* p) {
  size_t l;
  int    i, j;
  char   t[MAX_PERM * N];

  if (write(out_file, p, sizeof(POSET)) != sizeof(POSET))
    fatal_error("put_poset, writing of the poset failed");
  for (i = 0; i < p->p; i++)
    for (j = 0; j < p->m; j++)
      t[i * p->m + j] = (char)new_perm[i][j];
  l = p->p * p->m;
  if (write(out_file, t, l) != l)
    fatal_error("put_poset, writing of permutations failed");
  return l + sizeof(POSET);
}

static void get_header(int file, FILE_HEADER *fh) {
  if (read(file, fh, sizeof(FILE_HEADER)) != sizeof(FILE_HEADER))
    fatal_error("get_header, reading file header failed");
}

static void update_header(FILE_HEADER *fh) {
  clock_t c;
  struct tms t;

  if (lseek(out_file, 0, SEEK_SET) < 0)
    fatal_error("update_header, lseek(0) failed");
  fh->isomorphism_calls = isomorphic_calls_number;
  fh->exhaustive_calls = exhaustive_calls_number;
  fh->cpcp_calls = cpcp_calls_number;
  c = times(&t);
  fh->real_time += c - start_time;
  start_time = c;
  fh->user_time = t.tms_utime + user_time;
  fh->syst_time = t.tms_stime + syst_time;
  if (write(out_file, fh, sizeof(FILE_HEADER)) != sizeof(FILE_HEADER))
    fatal_error("update_header, writing file header failed");
  if (lseek(out_file, 0, SEEK_END) < 0)
    fatal_error("update_header, lseek(end) failed");
}

/*
 * Hash table routines
 */

static void empty_hash_table(void) {
  INDEX i;

  for (i = 0; i < HASH_TBL_SIZE; i++)
    hash_tbl[i] = EMPTY_POS;
}

static INDEX search(POSET *g) {
  INDEX h;
  int   i;

  for (h = 0, i = 0; i < g->n; i++)
    h += h1_tbl[cardinality(g->r[1][i])][cardinality(g->r[0][i])];
  h %= HASH_TBL_SIZE;
  g->hash = h;
  for (h = hash_tbl[h]; h != EMPTY_POS; h = posets_stack[h].hash)
    if (posets_stack[h].n == g->n)
      if (isomorphic(posets_stack + h, g))
        return h;
  return EMPTY_POS;
}

static void insert(const POSET *g) {
  if (new_posets_count >= MAX_STACK)
    fatal_error("insert, posets stack overflow");
  posets_stack[new_posets_count] = *g;
  posets_stack[new_posets_count].hash = hash_tbl[g->hash];
  hash_tbl[g->hash] = new_posets_count;
  new_posets_count++;
}

/*
 * Poset auxiliary routines
 */

static void poset_add_pair(POSET *g, int v1, int v2) {
  int v;
  SET v_set, n_set;

  /* Add pair (v1,v2) to r[0]. All elements till now greater than v2 are now
     greater than v1. If v was lower than v1, then now v is lower than all
     new elements greater than v1. */
  n_set = g->r[0][v1] |= pof2(v2) | g->r[0][v2];
  v_set = g->r[1][v1];
  while (v_set) {
    v = min_element(v_set);
    g->r[0][v] |= n_set;
    v_set ^= pof2(v);
  }

  /* Add pair (v2,v1) to r[1]. All elements till now lower than v1 are now
     lower than v2. If v was greater than v2, then now v is greater than all
     new elements lower than v2. */
  n_set = g->r[1][v2] |= pof2(v1) | g->r[1][v1];
  v_set = g->r[0][v2];
  while (v_set) {
    v = min_element(v_set);
    g->r[1][v] |= n_set;
    v_set ^= pof2(v);
  }
}

/*
 * Proof routines - searching for witness
 */

#ifdef _PRINT_LOG_
static long le_counter;

void le_print(const int e[], int n) {
  int i;

  printf("linear extension %ld:", le_counter);
  for (i = 0; i < n; i++)
    printf(" %d", e[i]);
  le_counter++;
}
#endif

static BOOL no_simple_witness(const POSET* p, const int e[], const int a[]) {
  int j, x, y, z;

  for (x = 0; x < p->m - 1; x++)
    for (y = x + 1; y < p->m; y++)
      if (notin(p->r[0][e[x]], e[y]) &&
          cardinality((p->r[0][e[x]] & ~p->r[0][e[y]]) |
                      (p->r[1][e[y]] & ~p->r[1][e[x]])) <= 1) {
        #ifdef _PRINT_LOG_
        printf(": (%d, %d) has at most one slave\n", e[x], e[y]);
        #endif
        lw2_x = a[x];
        lw2_y = a[y];
        return FALSE;
      }
  for (j = 2; j < p->m; j++)
    for (x = 0, z = x + j; z < p->m; x++, z++)
      for (y = x + 1; y < z; y++)
        if ((notin(p->r[0][e[x]], e[y]) &&
             (p->r[0][e[x]] & ~(p->r[0][e[y]] | p->r[0][e[z]] | pof2(e[z]))) ==
              EMPTY_SET && (p->r[1][e[y]] & ~p->r[1][e[x]]) == EMPTY_SET) ||
            (notin(p->r[1][e[z]], e[y]) &&
             (p->r[1][e[z]] & ~(p->r[1][e[y]] | p->r[1][e[x]] | pof2(e[x]))) ==
              EMPTY_SET && (p->r[0][e[y]] & ~p->r[0][e[z]]) == EMPTY_SET)) {
          #ifdef _PRINT_LOG_
          printf(": (%d, %d, %d) is balanced\n", e[x], e[y], e[z]);
          #endif
          lw3_x = a[x];
          lw3_y = a[y];
          lw3_z = a[z];
          return FALSE;
        }
  return TRUE;
}

static CPC t2(int pidx, int j, int k) {
  POSET p;

  if (t2cache[pidx][j][k] < 0) {
    if (in(pxyz[pidx].r[0][j], k)) {      /* j below k */
      t2cache[pidx][j][k] = txyz[pidx];
      t2cache[pidx][k][j] = 0;
    }
    else if (in(pxyz[pidx].r[1][j], k)) { /* j above k */
      t2cache[pidx][j][k] = 0;
      t2cache[pidx][k][j] = txyz[pidx];
    }
    else {                                /* j incomparable with k */
      p = pxyz[pidx];
      if (cardinality(p.r[1][j]) + cardinality(p.r[0][k]) >=
          cardinality(p.r[0][j]) + cardinality(p.r[1][k])) {
        poset_add_pair(&p, j, k);
        t2cache[pidx][k][j] = txyz[pidx] - (t2cache[pidx][j][k] = cpcp(&p));
      }
      else {
        poset_add_pair(&p, k, j);
        t2cache[pidx][j][k] = txyz[pidx] - (t2cache[pidx][k][j] = cpcp(&p));
      }
    }
  }
  return t2cache[pidx][j][k];
}

static CPC t3(int pidx, int x, int y, int z) {
  POSET p_tmp;

  if (in(pxyz[pidx].r[0][y], x) || /* y below z */
      in(pxyz[pidx].r[0][z], y) || /* z below y */
      in(pxyz[pidx].r[0][z], x))   /* z below x */
    return 0;
  p_tmp = pxyz[pidx];
  if (notin(pxyz[pidx].r[0][x], y)) /* x incomparable with y */
    poset_add_pair(&p_tmp, x, y);
  if (notin(pxyz[pidx].r[0][y], z)) /* y incomparable with z */
    poset_add_pair(&p_tmp, y, z);
  return cpcp(&p_tmp);
}

static void decomposition(const POSET* p, int l) {
  POSET p1;
  int   j;
  SET   L;

  while (l < p->n && (L = ~(p->r[1][l]) & p->M) == EMPTY_SET)
    l++;
  if (l >= p->n) {
    n_components++;
    if (n_components > MAX_COMPONENTS)
      fatal_error("decomposition, too many components");
    pxyz[n_components] = *p;
  }
  else {
    j = min_element(L);
    p1 = *p;
    poset_add_pair(&p1, l, j);
    p1.M |= pof2(l);
    decomposition(&p1, l + 1);
    p1 = *p;
    poset_add_pair(&p1, j, l);
    decomposition(&p1, l);
  }
}

static BOOL no_advanced_witness(const POSET* p, const int e[], const int a[]) {
  CPC  t_xyz;
  BOOL f1, f2;
  int  i, j, x, y, z;
  #ifdef _CHECK_CORRECTNESS_
  SET  H;
  #endif

  if (cache_not_valid) {
    n_components = 0;
    decomposition(p, p->m);
    for (i = 1; i <= n_components; i++)
      txyz[i] = cpcp(pxyz + i);
    #ifdef _CHECK_CORRECTNESS_
    /* Check decomposition consistency. */
    for (i = 1; i <= n_components; i++) {
      H = (FULL_SET >> (N - pxyz[i].n)) & ~pxyz[i].M;
      while (H) {
        x = min_element(H);
        if (~pxyz[i].r[1][x] & pxyz[i].M) {
          poset_print(pxyz + i, NULL, 0);
          fatal_error("no_advanced_witness, "
                      "decomposition failed for element %d", x);
        }
        H ^= pof2(x);
      }
    }
    #endif
    /* Reset t2 cache. Position 0 is currently not used! */
    for (i = 1; i <= n_components; i++)
      for (x = 0; x < p->m; x++)
        for (y = 0; y < p->m; y++)
          t2cache[i][x][y] = -1;
    cache_not_valid = FALSE;
  }
  for (x = 0; x < p->m - 1; x++) {
    for (y = x + 1; y < p->m; y++) {
      for (i = 1; i <= n_components && 2 * t2(i, e[x], e[y]) <= txyz[i]; i++);
      if (i > n_components) {
        #ifdef _PRINT_LOG_
        printf(": e(P + %d %d) <= 1/2 * e(P)\n", e[x], e[y]);
        #endif
        lw2_x = a[x];
        lw2_y = a[y];
        return FALSE;
      }
    }
  }
  for (j = 2; j < p->m; j++) {
    for (x = 0, z = x + j; z < p->m; x++, z++) {
      for (y = x + 1; y < z; y++) {
        i = 1;
        f1 = f2 = TRUE;
        while ((f1 || f2) && i <= n_components) {
          t_xyz = t3(i, e[x], e[y], e[z]);
          if (f1 && t_xyz > t2(i, e[y], e[x]))
            f1 = FALSE;
          if (f2 && t_xyz > t2(i, e[z], e[y]))
            f2 = FALSE;
          i++;
        }
        if (f1) {
          #ifdef _PRINT_LOG_
          printf(": e(P + %d %d %d) <= e(P + %d %d)\n",
                 e[x], e[y], e[z], e[y], e[x]);
          #endif
          lw3_x = a[x];
          lw3_y = a[y];
          lw3_z = a[z];
          return FALSE;
        }
        if (f2) {
          #ifdef _PRINT_LOG_
          printf(": e(P + %d %d %d) <= e(P + %d %d)\n",
                 e[x], e[y], e[z], e[z], e[y]);
          #endif
          lw3_x = a[x];
          lw3_y = a[y];
          lw3_z = a[z];
          return FALSE;
        }
      }
    }
  }
  return TRUE;
}

static BOOL no_witness(POSET* p, int m) {
  POSET p1;
  int   a_norm[N], a_prim[N], i, j, k, l, perms_to_check;

  perms_to_check = p->p;
  cache_not_valid = TRUE;
  p->p = 0;
  for (i = 0; i < perms_to_check; i++) {
    p1 = *p;
    for (j = 0; j < m - 1; j++)
      if (notin(p1.r[0][old_perm[i][j]], old_perm[i][j + 1]))
        poset_add_pair(&p1, old_perm[i][j], old_perm[i][j + 1]);
    /*
    for (j = 0; j < p1.m - 1; j++) {
      for (k = j + 1; k < p1.m; k++) {
        if (notin(p1.r[0][j], k) &&
            cardinality((p->r[0][k] & ~p->r[0][j]) |
                        (p->r[1][j] & ~p->r[1][k])) <= 1) {
          #ifdef _PRINT_LOG_
          printf("(%d, %d) has at most one slave\n", k, j);
          #endif
          poset_add_pair(&p1, j, k);
        }
      }
    }
    */
    lw2_x = lw2_y = lw3_x = lw3_y = lw3_z = 0;
    for (j = 0; j < p->m; j++)
      a_norm[j] = a_prim[j] = j;
V2:
    if (a_prim[lw2_x] < a_prim[lw2_y] ||
        (a_prim[lw3_x] < a_prim[lw3_y] && a_prim[lw3_y] < a_prim[lw3_z])) {
      #ifdef _PRINT_LOG_
      if (p->p >= MAX_PERM)
        fatal_error("no_witness, too many permutations");
      for (j = 0; j < p->m; j++)
        new_perm[p->p][j] = old_perm[i][a_norm[j]];
      for (j = p->m; j < p->n; j++)
        new_perm[p->p][j] = j;
      le_print(new_perm[p->p], p->m);
      printf(": as above\n");
      #endif
    }
    else {
      if (p->p >= MAX_PERM)
        fatal_error("no_witness, too many permutations");
      for (j = 0; j < p->m; j++)
        new_perm[p->p][j] = old_perm[i][a_norm[j]];
      for (j = p->m; j < p->n; j++)
        new_perm[p->p][j] = j;
      #ifdef _PRINT_LOG_
      le_print(new_perm[p->p], p->m);
      #endif
      if (no_simple_witness(p, new_perm[p->p], a_norm) &&
          no_advanced_witness(p, new_perm[p->p], a_norm)) {
        #ifdef _PRINT_LOG_
        printf(": NO WITNESS\n");
        #endif
        p->p++;
      }
    }
    k = p->m - 1;
V3:
    j = a_prim[k];
    /* if (j == 0 || in(p1.r[0][a_norm[j - 1]], k)) { */
    if (j == 0 || in(p1.r[0][old_perm[i][a_norm[j - 1]]], old_perm[i][k])) {
      while (j < k) {
        l = a_norm[j + 1];
        a_norm[j] = l;
        a_prim[l] = j;
        j++;
      }
      a_norm[k] = a_prim[k] = k;
      k--;
      if (k == 0)
        goto VEND;
    }
    else {
      l = a_norm[j - 1];
      a_norm[j - 1] = k;
      a_norm[j] = l;
      a_prim[k] = j - 1;
      a_prim[l] = j;
      goto V2;
    }
    goto V3;
VEND: ;
  }
  return p->p != 0;
}

/*
 * Proof main routines
 */

int first_configuration() {
  FILE_HEADER fh;
  POSET       p;
  int         i;

  empty_file_header(&fh);
  fh.max_elements      = 2;
  fh.output_posets     = 1;
  fh.considered_posets = 1;
  update_header(&fh);

  poset_empty(&p);
  p.m = 0;
  p.n = 2;
  p.p = 1;
  for (i = 0; i < N; i++)
    new_perm[0][i] = i;
  fh.file_length += put_poset(&p);
  update_header(&fh);

  print_file_header(&fh);
  return fh.output_posets == 0;
}

int one_step(BOOL recovery, long offset, long count) {
  FILE_HEADER in_fh, out_fh;
  POSET       *p0, p1;
  long        l;
  INDEX       from, to;
  int         a, c[K + 2], d[N + 2], i, j, m, q, r, u;
  SET         A, B;

  get_header(in_file, &in_fh);
  if (in_fh.thinness != K)
    fatal_error("one_step, missing thinness %d", in_fh.thinness);
  if (in_fh.version != VERSION)
    fatal_error("one_step, missing version %d", in_fh.version);

  if (recovery) {
    get_header(out_file, &out_fh);
    if (out_fh.thinness != K || out_fh.version != VERSION ||
        in_fh.step + 1 != out_fh.step)
      fatal_error("one_step recovery, missing output file, "
                  "thinnes %d, version %d, step %d",
                   out_fh.thinness, out_fh.version, out_fh.step);
    if (ftruncate(out_file, out_fh.file_length) < 0)
      fatal_error("one_step recovery, ftruncate failed");
    if (lseek(out_file, 0, SEEK_END) != out_fh.file_length)
      fatal_error("one_step recovery, lseek failed");
    user_time = out_fh.user_time;
    syst_time = out_fh.syst_time;
    isomorphic_calls_number = out_fh.isomorphism_calls;
    exhaustive_calls_number = out_fh.exhaustive_calls;
    cpcp_calls_number       = out_fh.cpcp_calls;
  }
  else {
    if (offset < 0)
      offset = 0;
    empty_file_header(&out_fh);
    out_fh.step = in_fh.step + 1;
    out_fh.input_offset = offset;
    update_header(&out_fh);
  }

  /* TO DO - skip posets from the input file */
  for (l = 0; l < out_fh.input_offset + out_fh.input_posets; l++)
    get_poset();

  c[0] = 0;  /* guardian */
  d[0] = -1; /* guardian */
  if (count <= 0 || count > in_fh.output_posets - out_fh.input_offset)
    count = in_fh.output_posets - out_fh.input_offset;
  while (out_fh.input_posets < count) {
    get_poset();
    m = posets_stack[0].m;
    posets_stack[0].m = posets_stack[0].n;
    if (posets_stack[0].n > out_fh.max_elements)
      out_fh.max_elements = posets_stack[0].n;
    new_posets_count = 1;
    to = 0;
    while (to < new_posets_count) {
      empty_hash_table();
      from = to;
      to = new_posets_count;
      for (p0 = posets_stack + from; p0 < posets_stack + to; p0++) {
        if (p0->n >= N)
          fatal_error("one_step, max number of elements exceeded");
        else if (p0->n - p0->m < 2*K - 2) {
          if (p0->n >= out_fh.max_elements)
            out_fh.max_elements = p0->n + 1;
          p0->M = FULL_SET >> (N - p0->m);
          /* d[1], ..., d[n] are elements incomparable with at most K-1 others.
             They are candidates to be incomparable with a new element. */
          for (u = 0, i = m; i < p0->n; i++)
            if (cardinality(p0->r[0][i] | p0->r[1][i]) >= p0->n - K)
              d[++u] = i;
          d[u + 1] = p0->n; /* guardian */
          for (r = u < K ? u : K; r > 0; r--) {
            /* c[1], ..., c[r] is r-combination from {1, 2, ..., n}. */
            for (i = 1; i <= r; i++)
              c[i] = i;
            c[r + 1] = u + 1; /* guardian */
            /* Oryginally was q = r; but it seems to be wrong when r = n. */
            q = r == u ? 1 : r;
            while (q >= 1 && d[c[1]] < p0->m) {
              p1 = *p0;
              p1.n++;
              out_fh.combinations++;
              A = EMPTY_SET;
              for (i = 0; i <= r; i++)
                for (j = d[c[i]] + 1; j < d[c[i + 1]]; j++)
                  /* poset_add_pair(&p1, j, p0->n); */
                  A |= p1.r[1][j] | pof2(j);
              /* if ((p1.r[1][p0->n] & p0->M) != p0->M) { */
              if ((A & p0->M) != p0->M) {
                p1.r[1][p0->n] = A;
                B = pof2(p0->n);
                while (A) {
                  a = min_element(A);
                  p1.r[0][a] |= B;
                  A ^= pof2(a);
                }
                out_fh.search_calls++;
                if (search(&p1) == EMPTY_POS)
                  insert(&p1);
              }
              /* Take the next combination in lexicografic order. */
              if (c[r] == u)
                q--;
              else
                q = r;
              if (q >= 1)
                for (i = r; i >= q; i--)
                  c[i] = c[q] + i - q + 1;
            }
          }
        }
      }
      if (new_posets_count - to > out_fh.hash_tbl_usage)
        out_fh.hash_tbl_usage = new_posets_count - to;
    }
    if (new_posets_count > out_fh.stack_usage)
      out_fh.stack_usage = new_posets_count;

    for (from = 0; from < new_posets_count; from++) {
      #ifdef _PRINT_LOG_
      le_counter = 0;
      printf("------------------------------------------------------------\n"
             "poset %ld  ", out_fh.considered_posets);
      poset_print(posets_stack + from, old_perm, m);
      #endif
      if (no_witness(posets_stack + from, m)) {
        if (posets_stack[from].m == posets_stack[from].n) {
          poset_print(posets_stack + from, new_perm, posets_stack[from].m);
          fatal_error("one_step, "
                      "configuration with empty separator not discarded");
        }
        out_fh.output_posets++;
        if (out_fh.max_permutations < posets_stack[from].p)
          out_fh.max_permutations = posets_stack[from].p;
        out_fh.file_length += put_poset(posets_stack + from);
      }
      if (n_components > out_fh.max_components)
        out_fh.max_components = n_components;
      if (!cache_not_valid)
        for (i = 1; i <= n_components; i++)
          if (txyz[i] > out_fh.max_extensions)
            out_fh.max_extensions = txyz[i];
      out_fh.considered_posets++;
    }
    out_fh.input_posets++;
    update_header(&out_fh);
  }
  #ifdef _PRINT_LOG_
  printf("------------------------------------------------------------\n");
  #endif
  print_file_header(&out_fh);
  return out_fh.output_posets == 0;
}

void print_files_contents(int argc, char* argv[], BOOL print_posets) {
  FILE_HEADER fh0, fh;
  long        i;
  int         k;

  empty_file_header(&fh0);
  fh0.step = -1;
  fh0.input_offset = (long)((unsigned long)(-1) >> 1);
  fh0.file_length = 0;
  for (k = 2; k < argc; k++) {
    in_file = open(argv[k], O_RDONLY);
    if (in_file < 0)
      fatal_error("print_files_contents, can't open file %s", argv[2]);
    get_header(in_file, &fh);
    if (fh.thinness != K)
      fatal_error("print_files_contents, missing thinness %d", fh.thinness);
    if (fh.version != VERSION)
      fatal_error("print_files_contents, missing version %d", fh.version);
    if (print_posets) {
      for (i = 0; i < fh.output_posets; i++) {
        get_poset();
        printf("poset %ld  ", i);
        poset_print(posets_stack, old_perm, posets_stack[0].m);
      }
    }
    close(in_file);
    compare_file_headers(&fh0, &fh);
  }
  in_file = -1;
  print_file_header(&fh0);
}
